import typer
from chroma_ops.collection_snapshot import command as snapshot_command

collection_commands = typer.Typer(no_args_is_help=True)

collection_commands.command(
    name="snapshot", no_args_is_help=True, help="Snapshot a collection to a file."
)(snapshot_command)



from typing import List, Tuple
import typer
import chromadb
from chroma_ops.constants import DEFAULT_TENANT_ID, DEFAULT_TOPIC_NAMESPACE
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    list_collections,
    validate_chroma_persist_dir,
)
from rich.console import Console
from rich.table import Table


def info_wal(persist_dir: str) -> List[Tuple[str, str, int]]:
    validate_chroma_persist_dir(persist_dir)
    console = Console()
    with get_sqlite_connection(persist_dir, SqliteMode.READ_ONLY) as conn:
        client = chromadb.PersistentClient(path=persist_dir)
        cursor = conn.cursor()
        all_collections = list_collections(client)
        collection_topics = {}
        stats = []
        for col in all_collections:
            collection_topics[
                f"persistent://{DEFAULT_TENANT_ID}/{DEFAULT_TOPIC_NAMESPACE}/{col.id}"
            ] = col.name
        query = "SELECT topic, COUNT(*) FROM embeddings_queue GROUP BY topic ORDER BY seq_id ASC;"
        res = cursor.execute(query).fetchall()
        table = Table(title="WAL Info")
        table.add_column("Collection")
        table.add_column("Topic")
        table.add_column("Count")
        for row in res:
            table.add_row(collection_topics[str(row[0])], str(row[0]), str(row[1]))
            stats.append((collection_topics[str(row[0])], str(row[0]), row[1]))
        console.print(table)
        return stats


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
) -> None:
    info_wal(persist_dir)



import typer

from chroma_ops.db_info import command as db_info_command
from chroma_ops.db_clean import command as db_clean_command

db_commands = typer.Typer(no_args_is_help=True)

db_commands.command(
    name="info",
    help="Provide persistent Chroma DB information. "
    "Useful to understand how your Chroma works or get support from the team.",
    no_args_is_help=True,
)(db_info_command)

db_commands.command(
    name="clean",
    help="Clean up orphanated HNSW segment subdirectories.",
    no_args_is_help=True,
)(db_clean_command)



import multiprocessing

DEFAULT_CHROMA_SQLITE_FILE = "chroma.sqlite3"
DEFAULT_TOPIC_NAMESPACE = "default"
DEFAULT_TENANT_ID = "default"
DEFAULT_DATABASE = "default_database"
DEFAULT_TENANT = "default_tenant"
DEFAULT_DISTANCE_METRIC = "l2"
DEFAULT_M = 16
DEFAULT_CONSTRUCTION_EF = 100
DEFAULT_SEARCH_EF = 10
DEFAULT_BATCH_SIZE = 100
DEFAULT_SYNC_THRESHOLD = 1000
DEFAULT_NUM_THREADS = multiprocessing.cpu_count()
DEFAULT_RESIZE_FACTOR = 1.2
DEFAULT_TOKENIZER = "trigram"



import base64
import json
import sys
from contextlib import contextmanager
from typing import Optional, Generator, IO, Union, Any, TextIO

import typer

from chroma_ops.constants import DEFAULT_TENANT_ID, DEFAULT_TOPIC_NAMESPACE
from chroma_ops.utils import (
    SqliteMode,
    get_sqlite_connection,
    validate_chroma_persist_dir,
)

from rich.console import Console
from rich.table import Table


@contextmanager
def smart_open(
    filename: Optional[str] = None,
) -> Generator[Union[IO[Any], TextIO], None, None]:
    fh: Union[IO[Any], TextIO] = sys.stdout
    if filename:
        fh = open(filename, "w")

    try:
        yield fh
    finally:
        if filename:
            fh.close()


def export_wal(
    persist_dir: str,
    output_file: str,
    *,
    tenant: Optional[str] = DEFAULT_TENANT_ID,
    topic_namespace: Optional[str] = DEFAULT_TOPIC_NAMESPACE,
    yes: Optional[bool] = False,
) -> None:
    validate_chroma_persist_dir(persist_dir)
    console = Console(stderr=True)
    table = Table(title="Exporting WAL")
    table.add_column("Collection", style="cyan")
    table.add_column("WAL Entries", style="magenta")
    with get_sqlite_connection(persist_dir, SqliteMode.READ_ONLY) as conn:
        collections = conn.execute(
            "SELECT c.name,c.id, s.id FROM collections c left join segments s on c.id=s.collection where s.scope='VECTOR'"
        ).fetchall()
        wal_topic_groups = conn.execute(
            "SELECT topic, count(*) FROM embeddings_queue group by topic"
        ).fetchall()
        for collection in collections:
            topic = f"persistent://{tenant}/{topic_namespace}/{collection[1]}"
            table.add_row(
                collection[0],
                str([s[1] for s in wal_topic_groups if s[0] == topic][0]),
            )
        console.print(table)
        if not yes:
            console.print("Are you sure you want to export the WAL? (y/N)")
            if not typer.confirm(
                "\nAre you sure you want to export the WAL?",
                default=False,
                show_default=True,
            ):
                console.print("[yellow]WAL export cancelled by user[/yellow]")
                return
        cursor = conn.cursor()
        query = "SELECT * FROM embeddings_queue ORDER BY seq_id ASC;"
        cursor.execute(query)
        exported_rows = 0
        with smart_open(output_file) as json_file:
            column_names = [description[0] for description in cursor.description]
            for row in cursor:
                row_data = {}
                for i, column_name in enumerate(column_names):
                    if column_name == "vector":
                        row_data[column_name] = base64.b64encode(row[i]).decode()
                    else:
                        row_data[column_name] = row[i]
                json_file.write(json.dumps(row_data) + "\n")
                exported_rows += 1

    console.print(f"[green]Exported {exported_rows} rows[/green]")


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    out: str = typer.Option(None, "--out", "-o", help="The output jsonl file"),
    yes: Optional[bool] = typer.Option(False, "--yes", "-y", help="Skip confirmation"),
) -> None:
    export_wal(persist_dir, out, yes=yes)



import os
from typing import Optional, Sequence, Any, Dict

import chromadb
import hnswlib
import typer
from chromadb.segment.impl.vector.hnsw_params import HnswParams
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from rich.rule import Rule
from rich.table import Table

from chroma_ops.utils import (
    SqliteMode,
    list_collections,
    get_sqlite_connection,
    validate_chroma_persist_dir,
    get_dir_size,
    decode_seq_id,
    PersistentData,
    sizeof_fmt,
    get_file_size,
)


def info(
    persist_dir: str,
    skip_collection_names: Optional[Sequence[str]] = None,
    privacy_mode: Optional[bool] = False,
) -> Dict[str, Any]:
    console = Console()
    validate_chroma_persist_dir(persist_dir)
    export_data = {}
    chroma_version = chromadb.__version__
    export_data["chroma_version"] = chroma_version
    persist_dir_size = sizeof_fmt(get_dir_size(persist_dir))
    export_data["persist_dir_size"] = persist_dir_size
    client = chromadb.PersistentClient(path=persist_dir)
    collection_count = client.count_collections()
    export_data["collection_count"] = collection_count
    collection_data = {}

    sql_file = os.path.join(persist_dir, "chroma.sqlite3")
    sysdb_size = sizeof_fmt(get_file_size(sql_file))
    export_data["sysdb_size"] = sysdb_size
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        active_hnsw_segment_dirs = set()
        console.print()
        with Progress(
            SpinnerColumn(finished_text="[bold green]:heavy_check_mark:[/bold green]"),
            TextColumn("[progress.description]{task.description}"),
            transient=True,
        ) as progress:
            task = progress.add_task(
                "Gathering collection data...", total=collection_count
            )
            for c in list_collections(client):
                if (
                    skip_collection_names is not None
                    and c.name in skip_collection_names
                ):
                    progress.update(task, advance=1)
                    continue
                collection = {
                    "id": str(c.id),
                    "name": c.name,
                    "metadata": c.metadata,
                    "tenant": c.tenant,
                    "database": c.database,
                    "records": c.count(),
                }
                cname = c.name
                cursor = conn.cursor()
                query = "SELECT s.*,c.* FROM segments s LEFT JOIN collections c ON s.collection = c.id WHERE c.id = ?;"
                cursor.execute(query, [collection["id"]])
                results = cursor.fetchall()
                collection["segments"] = []
                if len(results) > 0:
                    collection["dimension"] = (
                        c._model.dimension if hasattr(c, "_model") else results[0][8]
                    )
                for row in results:
                    segment = {
                        "id": row[0],
                        "type": row[1],
                        "scope": row[2],
                        "path": sql_file
                        if row[2] == "METADATA"
                        else os.path.join(persist_dir, row[0]),
                        "segment_metadata_path": None
                        if row[2] == "METADATA"
                        else os.path.join(persist_dir, row[0], "index_metadata.pickle"),
                    }
                    active_hnsw_segment_dirs.add(segment["path"])
                    max_seq_id_query = (
                        "SELECT seq_id FROM max_seq_id WHERE segment_id = ?"
                    )
                    cursor.execute(max_seq_id_query, [row[0]])  # N+1 query fun
                    results = cursor.fetchall()
                    segment["sysdb_max_seq_id"] = (
                        decode_seq_id(results[0][0]) if len(results) > 0 else 0
                    )
                    if segment["scope"] == "VECTOR":
                        segment["hnsw_dir_size"] = sizeof_fmt(
                            get_dir_size(segment["path"])
                        )
                        segment["hnsw_metadata_max_seq_id"] = 0
                        segment["hnsw_metadata_total_elements"] = 0
                        segment["wal_gap"] = 0
                        segment["hnsw_raw_total_elements"] = 0
                        segment["hnsw_orphan_elements"] = set()
                        segment["fragmentation_level"] = 0.0
                        segment["hnsw_raw_capacity"] = 0
                        if os.path.exists(segment["segment_metadata_path"]):
                            hnsw_metadata = PersistentData.load_from_file(
                                segment["segment_metadata_path"]
                            )
                            # support chroma 0.5.7+
                            if hasattr(hnsw_metadata, "max_seq_id"):
                                segment[
                                    "hnsw_metadata_max_seq_id"
                                ] = hnsw_metadata.max_seq_id
                            else:
                                max_seq_id_query_hnsw_057 = (
                                    "SELECT seq_id FROM max_seq_id WHERE segment_id = ?"
                                )
                                cursor.execute(max_seq_id_query_hnsw_057, [row[0]])
                                results = cursor.fetchall()
                                segment["hnsw_metadata_max_seq_id"] = (
                                    decode_seq_id(results[0][0])
                                    if len(results) > 0
                                    else 0
                                )
                            segment["hnsw_metadata_total_elements"] = len(
                                hnsw_metadata.id_to_label
                            )
                            segment["wal_gap"] = collection["records"] - len(
                                hnsw_metadata.id_to_label
                            )
                            hnsw_params = HnswParams(collection["metadata"])
                            index = hnswlib.Index(
                                space=hnsw_params.space, dim=collection["dimension"]
                            )
                            index.load_index(
                                os.path.join(segment["path"]),
                                is_persistent_index=True,
                                max_elements=segment["hnsw_metadata_max_seq_id"],
                            )
                            hnsw_ids = index.get_ids_list()
                            segment["hnsw_raw_total_elements"] = len(hnsw_ids)
                            segment["hnsw_raw_capacity"] = index.element_count
                            # fragmentation ration tells us the ratio of active elements vs total elements -
                            # the greater this number the more fragmented the index is impacting memory and performance
                            segment["fragmentation_level"] = (
                                (
                                    index.element_count
                                    - segment["hnsw_raw_total_elements"]
                                )
                                / index.element_count
                            ) * 100
                            hnsw_metadata_labels = set(hnsw_metadata.label_to_id.keys())
                            segment["hnsw_orphan_elements"] = (
                                set(hnsw_ids) - hnsw_metadata_labels
                            )
                            index.close_file_handles()
                    else:
                        # TODO implement stats on the metadata segment
                        # metadata segment
                        # metadata_info_query = "select count(*) from embedding_metadata  where segment_id = ? AND key <> 'chroma:document' group by id;"
                        ...
                    collection["segments"].append(segment)
                # Get WAL entries per collection
                query = "SELECT count(*) FROM embeddings_queue WHERE topic = ?;"
                cursor.execute(
                    query, [f"persistent://default/default/{collection['id']}"]
                )
                results = cursor.fetchall()
                collection["wal_entries"] = results[0][0]
                collection_data[cname] = collection
                progress.update(task, advance=1)

    # list dirs under persist_dir
    orphan_hnsw_dirs = []
    for entry in os.scandir(persist_dir):
        if (
            entry.is_dir()
            and os.path.join(persist_dir, entry.name) not in active_hnsw_segment_dirs
        ):
            orphan_hnsw_index = {
                "size": sizeof_fmt(get_dir_size(entry.path)),
                "path": entry.path if not privacy_mode else "redacted",
                "id": entry.name,
            }
            orphan_hnsw_dirs.append(orphan_hnsw_index)

    gtable = Table(title="General Info", expand=True)
    gtable.add_column("Property", justify="right", style="cyan", no_wrap=True)
    gtable.add_column("Value", style="magenta")
    gtable.add_row("Chroma Version", str(chroma_version))
    gtable.add_row("Number of Collection", str(collection_count))
    gtable.add_row(
        "Persist Directory", str(persist_dir) if not privacy_mode else "redacted"
    )
    gtable.add_row("Persist Directory Size", str(persist_dir_size))
    gtable.add_row(
        "SystemDB size: ",
        f"{sysdb_size} ({sql_file if not privacy_mode else 'redacted'})",
    )
    gtable.add_row("Orphan HNSW Directories", str(orphan_hnsw_dirs))
    console.print(gtable)
    console.print(Rule("Collections", style="red"))
    # Creating a table for the main information
    for cname, data in collection_data.items():
        tname = cname if not privacy_mode else data["id"]
        table = Table(title=f"'{tname}' Collection Data", expand=True)

        table.add_column("Table Data", justify="right", style="cyan", no_wrap=True)
        table.add_column("Value", style="magenta")

        table.add_row("ID", data["id"])
        table.add_row("Name", data["name"] if not privacy_mode else "redacted")
        table.add_row(
            "Metadata",
            str(data["metadata"])
            if not privacy_mode
            else str(
                {k: v for k, v in data["metadata"].items() if k.startswith("hnsw:")}
                if data["metadata"]
                else None
            ),
        )
        table.add_row("Dimension", str(data["dimension"]))
        table.add_row("Tenant", data["tenant"])
        table.add_row("Database", data["database"])
        table.add_row("Records", f"{data['records']:,}")
        table.add_row("WAL Entries", f"{data['wal_entries']:,}")

        metadata_segment = [
            segment for segment in data["segments"] if segment["scope"] == "METADATA"
        ][0]

        metadata_segment_table = Table(title=f"Metadata Segment ({tname})", expand=True)

        metadata_segment_table.add_column(
            "Property", justify="right", style="cyan", no_wrap=True
        )
        metadata_segment_table.add_column("Value", style="magenta")

        metadata_segment_table.add_row("Segment ID", metadata_segment["id"])
        metadata_segment_table.add_row("Type", metadata_segment["type"])
        metadata_segment_table.add_row("Scope", metadata_segment["scope"])
        metadata_segment_table.add_row(
            "SysDB Max Seq ID", f"{metadata_segment['sysdb_max_seq_id']:,}"
        )

        hnsw_segment = [
            segment for segment in data["segments"] if segment["scope"] == "VECTOR"
        ][0]

        hnsw_segment_table = Table(title=f"HNSW Segment ({tname})", expand=True)

        hnsw_segment_table.add_column(
            "Property", justify="right", style="cyan", no_wrap=True
        )
        hnsw_segment_table.add_column("Value", style="magenta")

        hnsw_segment_table.add_row("Segment ID", hnsw_segment["id"])
        hnsw_segment_table.add_row("Type", hnsw_segment["type"])
        hnsw_segment_table.add_row("Scope", hnsw_segment["scope"])
        hnsw_segment_table.add_row(
            "Path",
            hnsw_segment["path"] if not privacy_mode else f"*/{hnsw_segment['id']}",
        )
        hnsw_segment_table.add_row(
            "SysDB Max Seq ID", f"{hnsw_segment['sysdb_max_seq_id']:,}"
        )
        hnsw_segment_table.add_row("HNSW Dir Size", hnsw_segment["hnsw_dir_size"])
        hnsw_segment_table.add_row(
            "HNSW Metadata Max Seq ID", f"{hnsw_segment['hnsw_metadata_max_seq_id']:,}"
        )
        hnsw_segment_table.add_row(
            "HNSW Metadata Total Labels",
            f"{hnsw_segment['hnsw_metadata_total_elements']:,}",
        )
        hnsw_segment_table.add_row("WAL Gap", f"{hnsw_segment['wal_gap']:,}")
        hnsw_segment_table.add_row(
            "HNSW Raw Total Active Labels",
            f"{hnsw_segment['hnsw_raw_total_elements']:,}",
        )
        hnsw_segment_table.add_row(
            "HNSW Raw Allocated Labels", f"{hnsw_segment['hnsw_raw_capacity']:,}"
        )
        hnsw_segment_table.add_row(
            "HNSW Orphan Labels", str(hnsw_segment["hnsw_orphan_elements"])
        )
        hnsw_segment_table.add_row(
            "Fragmentation Level", str(hnsw_segment["fragmentation_level"])
        )
        console.print(Rule(f"{tname}", style="yellow"))
        console.print(table)
        console.print(Rule("Segments"))
        console.print(metadata_segment_table)
        console.print(hnsw_segment_table)
    export_data["collections"] = collection_data
    return export_data


def command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    skip_collection_names: str = typer.Option(
        None,
        "--skip-collection-names",
        "-s",
        help="Comma separated list of collection names to skip",
    ),
    privacy_mode: bool = typer.Option(
        False,
        "--privacy-mode",
        "-p",
        help="Redact sensitive data such as paths and names",
    ),
) -> None:
    info(
        persist_dir,
        skip_collection_names=skip_collection_names,
        privacy_mode=privacy_mode,
    )



import datetime
import json
import os
import shutil
import sqlite3
import tempfile
import traceback
from typing import Any, Callable, Dict, List, Optional, Tuple, TypedDict
import hnswlib
import typer
from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn
from rich.table import Table
from chroma_ops.utils import (
    DistanceMetric,
    SqliteMode,
    check_disk_space,
    get_sqlite_connection,
    validate_chroma_persist_dir,
    get_dir_size,
    PersistentData,
    sizeof_fmt,
)
from chroma_ops.constants import (
    DEFAULT_BATCH_SIZE,
    DEFAULT_CONSTRUCTION_EF,
    DEFAULT_DISTANCE_METRIC,
    DEFAULT_M,
    DEFAULT_NUM_THREADS,
    DEFAULT_RESIZE_FACTOR,
    DEFAULT_SEARCH_EF,
    DEFAULT_SYNC_THRESHOLD,
)

hnsw_commands = typer.Typer(no_args_is_help=True)


class HnswDetails(TypedDict):
    collection_name: str
    database: str
    space: str
    dimensions: int
    construction_ef: int
    search_ef: int
    m: int
    num_threads: int
    resize_factor: float
    batch_size: int
    sync_threshold: int
    segment_id: str
    path: str
    has_metadata: bool
    num_elements: int
    id_to_label: Dict[str, int]
    collection_id: str
    index_size: int
    fragmentation_level: float
    fragmentation_level_estimated: bool


def _get_hnsw_details(
    conn: sqlite3.Connection,
    persist_dir: str,
    collection_name: str,
    database: Optional[str] = "default_database",
    verbose: Optional[bool] = False,
) -> HnswDetails:
    collection_details = conn.execute(
        "SELECT id,dimension FROM collections WHERE name = ?", (collection_name,)
    ).fetchone()
    # config = json.loads(collection_details[1])
    config = {}

    segment_id = conn.execute(
        "SELECT id FROM segments WHERE scope = 'VECTOR' AND collection = ?",
        (collection_details[0],),
    ).fetchone()
    segment_metadata = conn.execute(
        """ SELECT json_object(
    'segment_id', segment_id,
    'metadata', json_object(
        'hnsw:search_ef',MAX(CASE WHEN key = 'hnsw:search_ef' THEN int_value END),
        'hnsw:construction_ef', MAX(CASE WHEN key = 'hnsw:construction_ef' THEN int_value END),
        'hnsw:M', MAX(CASE WHEN key = 'hnsw:M' THEN int_value END),
        'hnsw:batch_size', MAX(CASE WHEN key = 'hnsw:batch_size' THEN int_value END),
        'hnsw:sync_threshold', MAX(CASE WHEN key = 'hnsw:sync_threshold' THEN int_value END),
        'hnsw:space', MAX(CASE WHEN key = 'hnsw:space' THEN str_value END),
        'hnsw:num_threads', MAX(CASE WHEN key = 'hnsw:num_threads' THEN int_value END),
        'hnsw:resize_factor', MAX(CASE WHEN key = 'hnsw:resize_factor' THEN float_value END)
        )
    ) AS result_json
    FROM segment_metadata WHERE segment_id = ?""",
        (segment_id[0],),
    ).fetchone()

    config = json.loads(segment_metadata[0])["metadata"]
    space = (
        config["hnsw:space"]
        if "hnsw:space" in config and config["hnsw:space"]
        else DEFAULT_DISTANCE_METRIC
    )
    construction_ef = (
        config["hnsw:construction_ef"]
        if "hnsw:construction_ef" in config and config["hnsw:construction_ef"]
        else DEFAULT_CONSTRUCTION_EF
    )
    search_ef = (
        config["hnsw:search_ef"]
        if "hnsw:search_ef" in config and config["hnsw:search_ef"]
        else DEFAULT_SEARCH_EF
    )
    m = config["hnsw:M"] if "hnsw:M" in config and config["hnsw:M"] else DEFAULT_M
    num_threads = (
        config["hnsw:num_threads"]
        if "hnsw:num_threads" in config and config["hnsw:num_threads"]
        else DEFAULT_NUM_THREADS
    )
    resize_factor = (
        config["hnsw:resize_factor"]
        if "hnsw:resize_factor" in config and config["hnsw:resize_factor"]
        else DEFAULT_RESIZE_FACTOR
    )
    batch_size = (
        config["hnsw:batch_size"]
        if "hnsw:batch_size" in config and config["hnsw:batch_size"]
        else DEFAULT_BATCH_SIZE
    )
    sync_threshold = (
        config["hnsw:sync_threshold"]
        if "hnsw:sync_threshold" in config and config["hnsw:sync_threshold"]
        else DEFAULT_SYNC_THRESHOLD
    )
    dimensions = collection_details[1]
    id_to_label = {}
    fragmentation_level = 0.0
    fragmentation_level_estimated = True

    if os.path.exists(
        os.path.join(persist_dir, segment_id[0], "index_metadata.pickle")
    ):
        has_metadata = True
        persistent_data = PersistentData.load_from_file(
            os.path.join(persist_dir, segment_id[0], "index_metadata.pickle")
        )
        id_to_label = persistent_data.id_to_label
        if len(id_to_label) > 0:
            fragmentation_level = (
                (persistent_data.total_elements_added - len(id_to_label))
                / persistent_data.total_elements_added
                * 100
            )
        else:
            fragmentation_level = 0.0
            fragmentation_level_estimated = False
        if verbose:
            index = hnswlib.Index(space=space, dim=dimensions)
            index.load_index(
                os.path.join(persist_dir, segment_id[0]),
                is_persistent_index=True,
                max_elements=len(id_to_label),
            )
            index.set_num_threads(num_threads)
            index.set_ef(search_ef)
            total_elements = index.element_count
            if total_elements > 0:
                fragmentation_level = (
                    (total_elements - len(id_to_label)) / total_elements * 100
                )
                fragmentation_level_estimated = False
            else:
                fragmentation_level = 0.0
                fragmentation_level_estimated = False
            index.close_file_handles()
    else:
        has_metadata = False

    return HnswDetails(
        collection_name=collection_name,
        database=database if database else "default_database",
        space=space,
        dimensions=dimensions,
        construction_ef=construction_ef,
        search_ef=search_ef,
        m=m,
        num_threads=num_threads,
        resize_factor=resize_factor,
        batch_size=batch_size,
        sync_threshold=sync_threshold,
        segment_id=segment_id[0],
        path=os.path.join(persist_dir, segment_id[0]),
        has_metadata=has_metadata,
        num_elements=len(id_to_label),
        id_to_label=id_to_label,
        collection_id=collection_details[0],
        index_size=get_dir_size(os.path.join(persist_dir, segment_id[0])),
        fragmentation_level=fragmentation_level,
        fragmentation_level_estimated=fragmentation_level_estimated,
    )


def print_hnsw_details(hnsw_details: HnswDetails) -> None:
    console = Console()
    table = Table(
        title=f"HNSW details for collection [red]{hnsw_details['collection_name']}[/red] in [red]{hnsw_details['database']}[/red] database"
    )

    table.add_column("Metric", style="cyan")
    table.add_column("Value", style="green")

    # Add rows for each detail
    table.add_row("Space", str(hnsw_details["space"]))
    table.add_row("Dimensions", str(hnsw_details["dimensions"]))
    table.add_row("EF Construction", str(hnsw_details["construction_ef"]))
    table.add_row("EF Search", str(hnsw_details["search_ef"]))
    table.add_row("M", str(hnsw_details["m"]))
    table.add_row("Number of threads", str(hnsw_details["num_threads"]))
    table.add_row("Resize factor", str(hnsw_details["resize_factor"]))
    table.add_row("Batch size", str(hnsw_details["batch_size"]))
    table.add_row("Sync threshold", str(hnsw_details["sync_threshold"]))
    table.add_row("Segment ID", hnsw_details["segment_id"])
    table.add_row("Path", hnsw_details["path"])
    table.add_row("Has metadata", str(hnsw_details["has_metadata"]))
    table.add_row("Number of elements", str(hnsw_details["num_elements"]))
    table.add_row("Collection ID", str(hnsw_details["collection_id"]))
    table.add_row("Index size", sizeof_fmt(hnsw_details["index_size"]))
    table.add_row(
        "Fragmentation level",
        f"{hnsw_details['fragmentation_level']:.2f}% {'(estimated)' if hnsw_details['fragmentation_level_estimated'] else ''}",
    )
    console.print(table)


def _prepare_hnsw_segment_config_changes(
    segment_id: str,
    hnsw_details: HnswDetails,
    *,
    space: Optional[str] = None,
    construction_ef: Optional[int] = None,
    search_ef: Optional[int] = None,
    m: Optional[int] = None,
    num_threads: Optional[int] = None,
    resize_factor: Optional[float] = None,
    batch_size: Optional[int] = None,
    sync_threshold: Optional[int] = None,
) -> Tuple[
    List[Callable[[sqlite3.Connection], None]],
    Dict[str, Dict[str, Any]],
    HnswDetails,
]:
    """Prepare the HNSW segment config changes in `segment_metadata`"""
    changes_callbacks = []
    changes_diff: Dict[str, Dict[str, Any]] = {}
    final_changes: HnswDetails = hnsw_details.copy()
    if space and space != hnsw_details["space"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, str_value) VALUES (?, 'hnsw:space', ?) ON CONFLICT (segment_id, key) DO UPDATE SET str_value = ?",
                (segment_id, space, space),
            )
        )
        changes_diff["hnsw:space"] = {
            "old": hnsw_details["space"],
            "new": space,
        }
        final_changes["space"] = space
    if construction_ef and construction_ef != hnsw_details["construction_ef"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, int_value) VALUES (?, 'hnsw:construction_ef', ?) ON CONFLICT (segment_id, key) DO UPDATE SET int_value = ?",
                (segment_id, construction_ef, construction_ef),
            )
        )
        changes_diff["hnsw:construction_ef"] = {
            "old": hnsw_details["construction_ef"],
            "new": construction_ef,
        }
        final_changes["construction_ef"] = construction_ef
    if search_ef and search_ef != hnsw_details["search_ef"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, int_value) VALUES (?, 'hnsw:search_ef', ?) ON CONFLICT (segment_id, key) DO UPDATE SET int_value = ?",
                (segment_id, search_ef, search_ef),
            )
        )
        changes_diff["hnsw:search_ef"] = {
            "old": hnsw_details["search_ef"],
            "new": search_ef,
        }
        final_changes["search_ef"] = search_ef
    if m and m != hnsw_details["m"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, int_value) VALUES (?, 'hnsw:M', ?) ON CONFLICT (segment_id, key) DO UPDATE SET int_value = ?",
                (segment_id, m, m),
            )
        )
        changes_diff["hnsw:M"] = {
            "old": hnsw_details["m"],
            "new": m,
        }
        final_changes["m"] = m
    if num_threads and num_threads != hnsw_details["num_threads"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, int_value) VALUES (?, 'hnsw:num_threads', ?) ON CONFLICT (segment_id, key) DO UPDATE SET int_value = ?",
                (segment_id, num_threads, num_threads),
            )
        )
        changes_diff["hnsw:num_threads"] = {
            "old": hnsw_details["num_threads"],
            "new": num_threads,
        }
        final_changes["num_threads"] = num_threads
    if resize_factor and resize_factor != hnsw_details["resize_factor"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, float_value) VALUES (?, 'hnsw:resize_factor', ?) ON CONFLICT (segment_id, key) DO UPDATE SET float_value = ?",
                (segment_id, resize_factor, resize_factor),
            )
        )
        changes_diff["hnsw:resize_factor"] = {
            "old": hnsw_details["resize_factor"],
            "new": resize_factor,
        }
        final_changes["resize_factor"] = resize_factor
    if batch_size and batch_size != hnsw_details["batch_size"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, int_value) VALUES (?, 'hnsw:batch_size', ?) ON CONFLICT (segment_id, key) DO UPDATE SET int_value = ?",
                (segment_id, batch_size, batch_size),
            )
        )
        changes_diff["hnsw:batch_size"] = {
            "old": hnsw_details["batch_size"],
            "new": batch_size,
        }
        final_changes["batch_size"] = batch_size
    if sync_threshold and sync_threshold != hnsw_details["sync_threshold"]:
        changes_callbacks.append(
            lambda conn: conn.execute(
                "INSERT INTO segment_metadata (segment_id, key, int_value) VALUES (?, 'hnsw:sync_threshold', ?) ON CONFLICT (segment_id, key) DO UPDATE SET int_value = ?",
                (segment_id, sync_threshold, sync_threshold),
            )
        )
        changes_diff["hnsw:sync_threshold"] = {
            "old": hnsw_details["sync_threshold"],
            "new": sync_threshold,
        }
        final_changes["sync_threshold"] = sync_threshold
    return changes_callbacks, changes_diff, final_changes


def _print_hnsw_segment_config_changes(changes_diff: Dict[str, Dict[str, Any]]) -> None:
    console = Console()
    table = Table(title="HNSW segment config changes")
    table.add_column("Config Key", style="cyan")
    table.add_column("Old", style="green")
    table.add_column("New", style="red")
    for key, value in changes_diff.items():
        table.add_row(key, str(value["old"]), str(value["new"]))
    console.print(table)


def rebuild_hnsw(
    persist_dir: str,
    *,
    collection_name: str,
    database: Optional[str] = "default_database",
    backup: Optional[bool] = True,
    yes: Optional[bool] = False,
    space: Optional[str] = None,
    construction_ef: Optional[int] = None,
    search_ef: Optional[int] = None,
    m: Optional[int] = None,
    num_threads: Optional[int] = None,
    resize_factor: Optional[float] = None,
    batch_size: Optional[int] = None,
    sync_threshold: Optional[int] = None,
) -> None:
    """Rebuilds the HNSW index"""
    validate_chroma_persist_dir(persist_dir)
    console = Console()
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        # lock the database to ensure no new data is added while we are rebuilding the index
        conn.execute("BEGIN EXCLUSIVE")
        try:
            hnsw_details = _get_hnsw_details(
                conn, persist_dir, collection_name, database
            )
            (
                changes_callbacks,
                changes_diff,
                final_changes,
            ) = _prepare_hnsw_segment_config_changes(
                hnsw_details["segment_id"],
                hnsw_details,
                space=space,
                construction_ef=construction_ef,
                search_ef=search_ef,
                m=m,
                num_threads=num_threads,
                resize_factor=resize_factor,
                batch_size=batch_size,
                sync_threshold=sync_threshold,
            )
            if not hnsw_details["has_metadata"] and len(changes_diff) == 0:
                console.print(
                    f"[red]Index metadata not found for segment {hnsw_details['segment_id']} and no config changes to make. No need to rebuild.[/red]"
                )
                return
            _space = final_changes["space"]
            construction_ef = final_changes["construction_ef"]
            search_ef = final_changes["search_ef"]
            _m = final_changes["m"]
            _num_threads = final_changes["num_threads"]
            _batch_size = final_changes["batch_size"]
            segment_id = final_changes["segment_id"]
            dimensions = final_changes["dimensions"]
            id_to_label = final_changes["id_to_label"]
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_persist_dir = os.path.join(temp_dir, segment_id)
                if not check_disk_space(
                    os.path.join(persist_dir, segment_id), temp_dir
                ):
                    console.print(
                        f"[red]Not enough space on in temp dir {temp_persist_dir} to copy index from {os.path.join(persist_dir, segment_id)}[/red]"
                    )
                    return
                shutil.copytree(os.path.join(persist_dir, segment_id), temp_persist_dir)
                target_index = hnswlib.Index(space=_space, dim=dimensions)
                target_index.init_index(
                    max_elements=len(id_to_label),
                    ef_construction=construction_ef,
                    M=_m,
                    is_persistent_index=True,
                    persistence_location=temp_persist_dir,
                )
                target_index.set_num_threads(_num_threads)
                target_index.set_ef(search_ef)
                source_index = hnswlib.Index(space=_space, dim=dimensions)
                source_index.load_index(
                    os.path.join(persist_dir, segment_id),
                    is_persistent_index=True,
                    max_elements=len(id_to_label),
                )
                source_index.set_num_threads(_num_threads)
                values = list(id_to_label.values())
                print_hnsw_details(final_changes)
                if len(changes_diff) > 0:
                    _print_hnsw_segment_config_changes(changes_diff)
                if not yes:
                    if not typer.confirm(
                        "\nAre you sure you want to rebuild this index?",
                        default=False,
                        show_default=True,
                    ):
                        console.print("[yellow]Rebuild cancelled by user[/yellow]")
                        return
                if len(changes_diff) > 0:
                    for callback in changes_callbacks:
                        callback(conn)
                with Progress(
                    SpinnerColumn(
                        finished_text="[bold green]:heavy_check_mark:[/bold green]"
                    ),
                    TextColumn("[progress.description]{task.description}"),
                    *[
                        BarColumn(),
                        TextColumn("{task.percentage:>3.0f}%"),
                    ],  # Add these columns
                    transient=True,
                ) as progress:
                    task = progress.add_task(
                        "Adding items to target index...", total=len(values)
                    )
                    for i in range(0, len(values), _batch_size):
                        items = source_index.get_items(ids=values[i : i + _batch_size])
                        target_index.add_items(items, values[i : i + _batch_size])
                        progress.update(task, advance=len(items))
                target_index.persist_dirty()
                target_index.close_file_handles()
                source_index.close_file_handles()
                if backup:
                    backup_target = os.path.join(
                        persist_dir,
                        f"{segment_id}_backup_{datetime.datetime.now().strftime('%Y%m%d%H%M%S')}",
                    )
                    shutil.move(os.path.join(persist_dir, segment_id), backup_target)
                    console.print(
                        f"[bold green]Backup of old index created at {backup_target}[/bold green]"
                    )
                else:
                    shutil.rmtree(os.path.join(persist_dir, segment_id))
                shutil.copytree(temp_persist_dir, os.path.join(persist_dir, segment_id))
            conn.commit()
            print_hnsw_details(
                _get_hnsw_details(
                    conn, persist_dir, collection_name, database, verbose=True
                )
            )
        except Exception:
            conn.rollback()
            console.print("[red]Failed to rebuild HNSW index[/red]")
            traceback.print_exc()
            raise


def info_hnsw(
    persist_dir: str,
    collection_name: str,
    database: Optional[str] = "default_database",
    verbose: Optional[bool] = False,
) -> HnswDetails:
    validate_chroma_persist_dir(persist_dir)
    console = Console()
    with get_sqlite_connection(persist_dir, SqliteMode.READ_ONLY) as conn:
        try:
            hnsw_details = _get_hnsw_details(
                conn, persist_dir, collection_name, database, verbose=verbose
            )
            print_hnsw_details(hnsw_details)
            return hnsw_details
        except Exception:
            console.print("[red]Failed to get HNSW details[/red]")
            traceback.print_exc()
            raise


def modify_runtime_config(
    persist_dir: str,
    collection_name: str,
    database: str,
    search_ef: Optional[int] = None,
    num_threads: Optional[int] = None,
    resize_factor: Optional[float] = None,
    batch_size: Optional[int] = None,
    sync_threshold: Optional[int] = None,
    yes: Optional[bool] = False,
) -> None:
    validate_chroma_persist_dir(persist_dir)
    console = Console()
    with get_sqlite_connection(persist_dir, SqliteMode.READ_WRITE) as conn:
        conn.execute("BEGIN EXCLUSIVE")
        try:
            hnsw_details = _get_hnsw_details(
                conn, persist_dir, collection_name, database
            )
            (
                changes_callbacks,
                changes_diff,
                _,
            ) = _prepare_hnsw_segment_config_changes(
                hnsw_details["segment_id"],
                hnsw_details,
                search_ef=search_ef,
                num_threads=num_threads,
                resize_factor=resize_factor,
                batch_size=batch_size,
                sync_threshold=sync_threshold,
            )
            if len(changes_diff) > 0:
                _print_hnsw_segment_config_changes(changes_diff)
            else:
                console.print(
                    "[bold green]No changes to apply (it is possible requested changes are identical to current config)[/bold green]"
                )
                return
            if not yes:
                if not typer.confirm(
                    "\nAre you sure you want to apply these changes?",
                    default=False,
                    show_default=True,
                ):
                    console.print("[yellow]Changes cancelled by user[/yellow]")
                    return
            for callback in changes_callbacks:
                callback(conn)
            conn.commit()
            console.print(
                "[bold green]HNSW index configuration modified successfully[/bold green]"
            )
        except Exception:
            conn.rollback()
            console.print("[red]Failed to modify HNSW index configuration[/red]")
            traceback.print_exc()
            raise


def rebuild_hnsw_command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    collection_name: str = typer.Option(
        ..., "--collection", "-c", help="The collection name"
    ),
    database: str = typer.Option(
        "default_database",
        "--database",
        "-d",
        help="The database name",
    ),
    backup: bool = typer.Option(
        True,
        "--backup",
        "-b",
        help="Backup the old index before rebuilding",
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Skip confirmation prompt",
    ),
    space: Optional[DistanceMetric] = typer.Option(
        None,
        "--space",
        help="The space to use for the index",
        case_sensitive=False,
    ),
    construction_ef: Optional[int] = typer.Option(
        None,
        "--construction-ef",
        help="The construction ef to use for the index",
        min=1,
    ),
    search_ef: Optional[int] = typer.Option(
        None,
        "--search-ef",
        help="The search ef to use for the index",
        min=1,
    ),
    m: Optional[int] = typer.Option(
        None,
        "--m",
        help="The m to use for the index",
        min=1,
    ),
    num_threads: Optional[int] = typer.Option(
        None,
        "--num-threads",
        help="The number of threads to use for the index",
        min=1,
    ),
    resize_factor: Optional[float] = typer.Option(
        None,
        "--resize-factor",
        help="The resize factor to use for the index",
        min=1.0,
    ),
    batch_size: Optional[int] = typer.Option(
        None,
        "--batch-size",
        help="The batch size to use for the index",
        min=2,
    ),
    sync_threshold: Optional[int] = typer.Option(
        None,
        "--sync-threshold",
        help="The sync threshold to use for the index",
        min=2,
    ),
) -> None:
    rebuild_hnsw(
        persist_dir,
        collection_name=collection_name,
        database=database,
        backup=backup,
        yes=yes,
        space=space.lower() if space else None,
        construction_ef=construction_ef,
        search_ef=search_ef,
        m=m,
        num_threads=num_threads,
        resize_factor=resize_factor,
        batch_size=batch_size,
        sync_threshold=sync_threshold,
    )


def info_hnsw_command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    collection_name: str = typer.Option(
        ..., "--collection", "-c", help="The collection name"
    ),
    database: str = typer.Option(
        "default_database",
        "--database",
        "-d",
        help="The database name",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        help="Verbose output",
    ),
) -> None:
    info_hnsw(persist_dir, collection_name, database, verbose)


def hnsw_modify_runtime_config_command(
    persist_dir: str = typer.Argument(..., help="The persist directory"),
    *,
    collection_name: str = typer.Option(
        ..., "--collection", "-c", help="The collection name"
    ),
    database: str = typer.Option(
        "default_database",
        "--database",
        "-d",
        help="The database name",
    ),
    search_ef: Optional[int] = typer.Option(
        None,
        "--search-ef",
        help="The search ef to use for the index",
        min=1,
    ),
    num_threads: Optional[int] = typer.Option(
        None,
        "--num-threads",
        help="The number of threads to use for the index",
        min=1,
    ),
    resize_factor: Optional[float] = typer.Option(
        None,
        "--resize-factor",
        help="The resize factor to use for the index",
        min=1.0,
    ),
    batch_size: Optional[int] = typer.Option(
        None,
        "--batch-size",
        help="The batch size to use for the index",
        min=2,
    ),
    sync_threshold: Optional[int] = typer.Option(
        None,
        "--sync-threshold",
        help="The sync threshold to use for the index",
        min=2,
    ),
    yes: bool = typer.Option(
        False,
        "--yes",
        "-y",
        help="Skip confirmation prompt",
    ),
) -> None:
    modify_runtime_config(
        persist_dir,
        collection_name,
        database,
        search_ef=search_ef,
        num_threads=num_threads,
        resize_factor=resize_factor,
        batch_size=batch_size,
        sync_threshold=sync_threshold,
        yes=yes,
    )


hnsw_commands.command(
    name="rebuild",
    help="Rebuild the HNSW index and update HNSW index configuration",
    no_args_is_help=True,
)(rebuild_hnsw_command)

hnsw_commands.command(
    name="info",
    help="Info about the HNSW index",
    no_args_is_help=True,
)(info_hnsw_command)

hnsw_commands.command(
    name="config",
    help="Modify the HNSW index configuration. This is a soft change that updates index configuration without rebuilding the index. The config changes are related to ",
    no_args_is_help=True,
)(hnsw_modify_runtime_config_command)
